body {
  background-color: #f2f2f2;
}

header {
  background-color: #8B0000;
  color: #fff;
  padding: 20px;
  text-align: center;
}

nav ul {
  list-style: none;
  margin: 0;
  padding: 0;
}

nav li {
  display: inline-block;
  margin: 0 10px;
}

nav li a {
  color: #fff;
  text-decoration: none;
  font-size: 18px;
}

nav li a:hover {
  color: #FFFACD;
}

main {
  display: flex;/* 将 main 置为 flex 容器 */
  flex-direction: column;/* 表示主轴是纵向 */
  justify-content: center;/* 子元素在主轴上居中 */
  align-items: center;/* 子元素在侧轴上居中 */
}

#taskMessage, #taskFlashMessage {
  resize: none;
}