DEFAULT_USER_AGENT = "Fenjing/0.1"

LITERAL = "literal"
UNSATISFIED = "unsatisfied"
ZERO = "zero"
POSITIVE_INTEGER = "positive_integer"
INTEGER = "integer"
STRING_STRING_CONCNAT = "string_string_concat"
STRING_PERCENT = "string_percent"
STRING_PERCENT_LOWER_C = "string_percent_lower_c"
STRING_UNDERLINE = "string_underline"
STRING_LOWERC = "string_lower_c"
STRING_MANY_PERCENT_LOWER_C = "string_many_percent_lower_c"
STRING = "string"
FORMULAR_SUM = "formular_sum"
ATTRIBUTE = "attribute"
ITEM = "item"
CLASS_ATTRIBUTE = "class_attribute"
CHAINED_ATTRIBUTE_ITEM = "chained_attribute_item"
EVAL_FUNC = "eval_func"
EVAL = "eval"
CONFIG = "config"
MODULE_OS = "module_os"
OS_POPEN_OBJ = "os_popen_obj"
OS_POPEN_READ = "os_popen_read"

GEN_TYPES = [
    "literal",
    "unsatisfied",
    "zero",
    "positive_integer",
    "integer",
    "string_string_concat",
    "string_percent",
    "string_percent_lower_c",
    "string_underline",
    "string_lower_c",
    "string_many_percent_lower_c",
    "string",
    "formular_sum",
    "attribute",
    "item",
    "class_attribute",
    "chained_attribute_item",
    "eval_func",
    "eval",
    "config",
    "module_os",
    "os_popen_obj",
    "os_popen_read",
]

CALLBACK_PREPARE_FULLPAYLOADGEN = "prepare_fullpayloadgen"
CALLBACK_GENERATE_FULLPAYLOAD = "generate_full_payload"
CALLBACK_GENERATE_PAYLOAD = "payload_gen"
CALLBACK_SUBMIT = "submit"
CALLBACK_TEST_FORM_INPUT = "test_form_input"

APICODE_OK = 200
APICODE_WRONG_INPUT = 401
